﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Flyweight
{
    
    //stan wewnetrzny - jeden model jest wspoldzielony przez kilku zolnierzy    
    class Model
    {
        String name;
        object[] vertices;
        object material;
        object texture;

        public Model(String modelFile)
        {
            this.name = modelFile;
            Console.WriteLine("Wczytuje dane modelu z pliku {0}.obj",modelFile);
        }

        public void draw()
        {
            Console.WriteLine("Rysuje model {0}", this.name);
        }

    }
 
    //bazowa klasa zolnierza, przechowuje stan zewnetrzny i stanowi nadbudowe dla stanu wewnetrznego (soldierModel)
    //ten stan zewnetrzny moglby byc tez skladowany w jakiejs tablicy
    //stan zewnetrzny czesto tez podaje sie stanowi wewnetrznemu jako parametr
    // w tym przypadku w klasie model moglbym umiescic draw sparametryzowane od jakiejs klasy SoldierState
    //tak naprawde jesli chodzi o pylek to wystepuje bardzo szerokie pole do interpretacji
    abstract class Soldier
    {
        protected Model soldierModel = null;
        
        private float x = 0.0f;
        private float y = 0.0f;

        
        public bool hasModel() { return !(this.soldierModel == null); }
        public void move(float dx, float dy) { this.x += dx; this.y += dy; }
        public void draw() 
        {
            if (!this.hasModel())
                throw new NullReferenceException("Model nie zostal ustawiony.");
            
            //operacje translacji,rotacji itd. dla danego zolnierza 
            this.soldierModel.draw();

        }
    }

    class Archer : Soldier
    {
        public Archer()
        {
            //podobnie jak wyzej - tak naprawde kwestia gustu gdzie sie pobiera wartosci ze slownika
            this.soldierModel = SoldierModelFactory.getSoldierModel("Archer");
        }

        public void shootArrow()
        {

        }
    }

    class YariAshigaru: Soldier
    {
        public YariAshigaru()
        {
            this.soldierModel = SoldierModelFactory.getSoldierModel("Yari");
        }

        public void stab()
        {

        }

    }

    class SoldierModelFactory
    {
        //implementacja fabryki rowniez jest dosc dowolna
        //najczesciej jest to slownik
        static Dictionary<String, Model> SoldiersModels = new Dictionary<string,Model>();


        //fabryka rownie dobrze moglaby zwracac od razu zolnierzy na ktorych by sie operowalo poprzez setX,setY i rysowalo
        public static Model getSoldierModel(String key)
        {

            if (SoldierModelFactory.SoldiersModels.Keys.Contains(key))
                return SoldierModelFactory.SoldiersModels[key];
            
            Model newModel = new Model(key);
            SoldierModelFactory.SoldiersModels.Add(key, newModel);
            return newModel;
            
        }
    }

    class Program
    {
        static void Main(string[] args)
        {

            List<Soldier> soldiers = new List<Soldier>();

            //tworze 20 lucznikow i 10 ashigaru z yari
            for (int i = 0; i < 20; i++)
                soldiers.Add(new Archer());
            
            for (int i = 0; i < 10; i++)
                soldiers.Add(new YariAshigaru());

            //nalezy zauwazyc ze ladowanie modelu zostalo wywolane tylko raz
            //kazdy z zolnierzy posiada jedynie referencje do bazowego modelu, ktorej NIE MOZE ZMIENIAC
            foreach (Soldier s in soldiers)
                s.draw();

            Console.ReadKey();
        }
    }
}
